#include "ruby.h"
#ifdef HAVE_RUBY_IO_H
#  include "ruby/io.h"
#else
#  include "rubyio.h"
#endif
#include <fcntl.h>
#include <assert.h>
#include <sys/uio.h>
#include <limits.h>
#include <alloca.h>

#if ! HAVE_RB_IO_T
#  define rb_io_t OpenFile
#endif

#ifdef GetReadFile
#  define FPTR_TO_FD(fptr) (fileno(GetReadFile(fptr)))
#else
#  if !HAVE_RB_IO_T || (RUBY_VERSION_MAJOR == 1 && RUBY_VERSION_MINOR == 8)
#    define FPTR_TO_FD(fptr) fileno(fptr->f)
#  else
#    define FPTR_TO_FD(fptr) fptr->fd
#  endif
#endif

static int my_fileno(VALUE io)
{
	rb_io_t *fptr;

	for (;;) {
		switch (TYPE(io)) {
		case T_FIXNUM: return NUM2INT(io);
		case T_FILE: {
			GetOpenFile(io, fptr);
			return FPTR_TO_FD(fptr);
		}
		default:
			io = rb_convert_type(io, T_FILE, "IO", "to_io");
			/* retry */
		}
	}
}
#ifndef HAVE_RB_THREAD_BLOCKING_REGION
/* partial emulation of the 1.9 rb_thread_blocking_region under 1.8 */
#  include <rubysig.h>
#  define RUBY_UBF_IO ((rb_unblock_function_t *)-1)
typedef void rb_unblock_function_t(void *);
typedef VALUE rb_blocking_function_t(void *);
static VALUE
rb_thread_blocking_region(
	rb_blocking_function_t *fn, void *data1,
	rb_unblock_function_t *ubf, void *data2)
{
	VALUE rv;

	assert(RUBY_UBF_IO == ubf && "RUBY_UBF_IO required for emulation");

	TRAP_BEG;
	rv = fn(data1);
	TRAP_END;

	return rv;
}
#endif /* ! HAVE_RB_THREAD_BLOCKING_REGION */

#ifndef RSTRING_PTR
#  define RSTRING_PTR(s) (RSTRING(s)->ptr)
#endif
#ifndef RSTRING_LEN
#  define RSTRING_LEN(s) (RSTRING(s)->len)
#endif
#ifndef RARRAY_PTR
#  define RARRAY_PTR(s) (RARRAY(s)->ptr)
#endif
#ifndef RARRAY_LEN
#  define RARRAY_LEN(s) (RARRAY(s)->len)
#endif

/*
 * Releases GVL only iff blocking I/O is used.
 * We'll trust programmers who use non-blocking I/O explicitly to
 * want the fastest possible performance without resorting to threads,
 * so releasing and them immediately reacquiring the GVL would be
 * a waste of time.
 */
static VALUE nb_io_run(rb_blocking_function_t *fn, void *data, unsigned flags)
{
	if (flags & SPLICE_F_NONBLOCK)
		return fn(data);
	return rb_thread_blocking_region(fn, data, RUBY_UBF_IO, 0);
}

struct splice_args {
	int fd_in;
	off_t *off_in;
	int fd_out;
	off_t *off_out;
	size_t len;
	unsigned flags;
};

static VALUE nogvl_splice(void *ptr)
{
	struct splice_args *a = ptr;

	return (VALUE)splice(a->fd_in, a->off_in, a->fd_out, a->off_out,
	                     a->len, a->flags);
}

/*
 * call-seq:
 *    IO.splice(fd_in, off_in, fd_out, off_out, len, flags) => integer
 *
 * Splice +len+ bytes from/to a pipe.  Either +fd_in+ or +fd_out+
 * MUST be a pipe.  +fd_in+ and +fd_out+ may BOTH be pipes as of
 * Linux 2.6.31 or later.
 *
 * +off_in+ and +off_out+ if non-nil may be used to
 * specify an offset for the non-pipe file descriptor.
 *
 * +flags+ may be a bitmask of the following flags:
 *
 * IO::Splice::F_MOVE, IO::Splice::F_NONBLOCK, IO::Splice::F_MORE
 *
 * Returns the number of bytes spliced.
 * Raises EOFError when +fd_in+ has reached end of file.
 * Raises Errno::EAGAIN if the IO::Splice::F_NONBLOCK flag is set
 * and the pipe has no data to read from or space to write to.  May
 * also raise Errno::EAGAIN if the non-pipe descriptor has no data
 * to read from or space to write to.
 *
 *     rd, wr = (pipe = IO.pipe).map { |io| io.fileno }
 *     src_io, dst_io = File.open("/path/to/src"), File.open("/path/to/dst")
 *     src, dst = src_io.fileno, dst_io.fileno
 *
 *     nr = IO.splice(src, nil, wr, nil, IO::Splice::PIPE_CAPA, 0)
 *     IO.splice(rd, nil, dst, nil, nr, 0)
 *
 * As splice never exposes buffers to userspace, it will not take
 * into account userspace buffering done by Ruby or stdio.  It is
 * also not subject to encoding/decoding filters under Ruby 1.9.
 *
 * See manpage for full documentation:
 * http://kernel.org/doc/man-pages/online/pages/man2/splice.2.html
 */
static VALUE my_splice(VALUE self,
	VALUE fd_in, VALUE off_in,
	VALUE fd_out, VALUE off_out,
	VALUE len, VALUE flags)
{
	off_t i, o;
	long n;
	struct splice_args a = {
		.off_in = NIL_P(off_in) ? NULL : (i = NUM2OFFT(off_in), &i),
		.off_out = NIL_P(off_out) ? NULL : (o = NUM2OFFT(off_out), &o),
		.fd_in = my_fileno(fd_in),
		.fd_out = my_fileno(fd_out),
		.len = (size_t)NUM2ULONG(len),
		.flags = NUM2UINT(flags),
	};

	n = (long)rb_thread_blocking_region(nogvl_splice, &a, RUBY_UBF_IO, 0);
	if (n == 0)
		rb_eof_error();
	if (n < 0)
		rb_sys_fail("splice");
	return LONG2NUM(n);
}

struct tee_args {
	int fd_in;
	int fd_out;
	size_t len;
	unsigned flags;
};

/* runs without GVL */
static VALUE nogvl_tee(void *ptr)
{
	struct tee_args *a = ptr;

	return (VALUE)tee(a->fd_in, a->fd_out, a->len, a->flags);
}

/*
 * call-seq:
 *   IO.tee(fd_in, fd_out, len, flags) => integer
 *
 * Copies up to +len+ bytes of data from +fd_in+ to +fd_out+.  +fd_in+
 * and +fd_out+ must both refer to pipe descriptors.  +fd_in+ and +fd_out+
 * may not be endpoints of the same pipe.
 *
 * +flags+ may be zero or IO::Splice::F_NONBLOCK
 * Other IO::Splice flags are currently unimplemented or have no effect.
 *
 * Returns the number of bytes duplicated if successful.
 * Raises EOFError when +fd_in+ is closed and emptied.
 * Raises Errno::EAGAIN when +fd_in+ is empty and/or +fd_out+ is full
 * and +flags+ contains IO::Splice::F_NONBLOCK
 *
 * See manpage for full documentation:
 * http://kernel.org/doc/man-pages/online/pages/man2/tee.2.html
 */
static VALUE my_tee(VALUE self,
	VALUE fd_in, VALUE fd_out,
	VALUE len, VALUE flags)
{
	long n;
	struct tee_args a = {
		.fd_in = my_fileno(fd_in),
		.fd_out = my_fileno(fd_out),
		.len = (size_t)NUM2ULONG(len),
		.flags = NUM2UINT(flags),
	};

	n = (long)nb_io_run(nogvl_tee, &a, a.flags);
	if (n == 0)
		rb_eof_error();
	if (n < 0)
		rb_sys_fail("tee");

	return LONG2NUM(n);
}

struct vmsplice_args {
	int fd;
        struct iovec *iov;
	unsigned long nr_segs;
	unsigned flags;
};

static VALUE nogvl_vmsplice(void *ptr)
{
	struct vmsplice_args *a = ptr;

	return (VALUE)vmsplice(a->fd, a->iov, a->nr_segs, a->flags);
}

/* this can't be a function since we use alloca() */
#define ARY2IOVEC(iov,iovcnt,expect,ary) \
do { \
	VALUE *cur; \
	struct iovec *tmp; \
	long n; \
	Check_Type(ary, T_ARRAY); \
	cur = RARRAY_PTR(ary); \
	n = RARRAY_LEN(ary); \
	if (n > IOV_MAX) \
		rb_raise(rb_eArgError, "array is larger than IOV_MAX"); \
	iov = tmp = alloca(sizeof(struct iovec) * n); \
	expect = 0; \
	iovcnt = n; \
	for (; --n >= 0; tmp++, cur++) { \
		Check_Type(*cur, T_STRING); \
		tmp->iov_base = RSTRING_PTR(*cur); \
		tmp->iov_len = RSTRING_LEN(*cur); \
		expect += tmp->iov_len; \
	} \
} while (0)

static void advance_vmsplice_args(struct vmsplice_args *a, long n)
{
	struct iovec *new_iov = a->iov;
	int i;

	/* skip over iovecs we've already written completely */
	for (i = 0; i < a->nr_segs; i++, new_iov++) {
		if (n == 0)
			break;
		/*
		 * partially written iov,
		 * modify and retry with current iovec in
		 * front
		 */
		if (new_iov->iov_len > (size_t)n) {
			VALUE base = (VALUE)new_iov->iov_base;

			new_iov->iov_len -= n;
			new_iov->iov_base = (void *)(base + n);
			break;
		}

		n -= new_iov->iov_len;
	}

	/* setup to retry without the already-written iovecs */
	a->nr_segs -= i;
	a->iov = new_iov;
}

/*
 * call-seq:
 *   IO.vmsplice(fd, string_array, flags) => integer
 *
 * Transfers an array of strings into the pipe descriptor given by fd.
 * +fd+ must be the writable end of a pipe.
 *
 * This may allow the kernel to avoid data copies in some cases.
 * but is (probably) of limited usefulness in Ruby.
 *
 * See manpage for full documentation:
 * http://kernel.org/doc/man-pages/online/pages/man2/vmsplice.2.html
 */
static VALUE my_vmsplice(VALUE self, VALUE fd, VALUE data, VALUE flags)
{
	long rv = 0;
	ssize_t left;
	struct vmsplice_args a;

	ARY2IOVEC(a.iov, a.nr_segs, left, data);
	a.fd = my_fileno(fd);
	a.flags = NUM2UINT(flags);

	for (;;) {
		long n = (long)nb_io_run(nogvl_vmsplice, &a, a.flags);

		if (n < 0) {
			if (errno == EAGAIN) {
				if (a.flags & SPLICE_F_NONBLOCK)
					rb_sys_fail("vmsplice");
				else if (rb_io_wait_writable(a.fd))
					continue;
				/* fall through on error */
			}
			/*
			 * unlikely to hit this case, return the
			 * already written bytes, we'll let the next
			 * write (or close) fail instead
			 */
			if (rv > 0)
				break;
			rb_sys_fail("vmsplice");
		}

		rv += n;
		left -= n;
		if (left == 0)
			break;
		advance_vmsplice_args(&a, n);
	}

	return LONG2NUM(rv);
}

void Init_io_splice_ext(void)
{
	VALUE mSplice = rb_define_module_under(rb_cIO, "Splice");

	rb_define_singleton_method(rb_cIO, "splice", my_splice, 6);
	rb_define_singleton_method(rb_cIO, "tee", my_tee, 4);
	rb_define_singleton_method(rb_cIO, "vmsplice", my_vmsplice, 3);

	/*
	 * Attempt to move pages instead of copying.  This is only a hint
	 * and support for it was removed in Linux 2.6.21 and has not been
	 * readded as of 2.6.30.
	 */
	rb_define_const(mSplice, "F_MOVE", UINT2NUM(SPLICE_F_MOVE));

	/*
	 * Do not block on pipe I/O.  This flag only affects the pipe(s)
	 * being spliced from/to and has no effect on the non-pipe
	 * descriptor (which requires non-blocking operation to be set
	 * explicitly).
	 *
	 * The non-blocking flag (O_NONBLOCK) on the pipe descriptors
	 * themselves are ignored by this family of functions, and
	 * using this flag is the only way to get non-blocking operation
	 * out of them.
	 */
	rb_define_const(mSplice, "F_NONBLOCK", UINT2NUM(SPLICE_F_NONBLOCK));

	/*
	 * Indicate that there may be more data coming into the outbound
	 * descriptor.  This can allow the kernel to avoid sending partial
	 * frames from sockets.  Currently only used with splice.
	 */
	rb_define_const(mSplice, "F_MORE", UINT2NUM(SPLICE_F_MORE));

	/*
	 * Only usable by vmsplice.  This flag probably not useful in the
	 * context of Ruby applications which cannot control alignment.
	 */
	rb_define_const(mSplice, "F_GIFT", UINT2NUM(SPLICE_F_GIFT));

	/*
	 * The maximum size of an atomic write to a pipe
	 * POSIX requires this to be at least 512 bytes.
	 * Under Linux, this is 4096 bytes.
	 */
	rb_define_const(mSplice, "PIPE_BUF", UINT2NUM(PIPE_BUF));
}
